@@ -57,13 +57,11 @@
 
 #define FL_SINGLETON    RBIMPL_CAST((VALUE)RUBY_FL_SINGLETON)            /**< @old{RUBY_FL_SINGLETON} */
 #define FL_WB_PROTECTED RBIMPL_CAST((VALUE)RUBY_FL_WB_PROTECTED)         /**< @old{RUBY_FL_WB_PROTECTED} */
-#define FL_PROMOTED0    RBIMPL_CAST((VALUE)RUBY_FL_PROMOTED0)            /**< @old{RUBY_FL_PROMOTED0} */
-#define FL_PROMOTED1    RBIMPL_CAST((VALUE)RUBY_FL_PROMOTED1)            /**< @old{RUBY_FL_PROMOTED1} */
+#define FL_PROMOTED     RBIMPL_CAST((VALUE)RUBY_FL_PROMOTED)             /**< @old{RUBY_FL_PROMOTED} */
 #define FL_FINALIZE     RBIMPL_CAST((VALUE)RUBY_FL_FINALIZE)             /**< @old{RUBY_FL_FINALIZE} */
 #define FL_TAINT        RBIMPL_CAST((VALUE)RUBY_FL_TAINT)                /**< @old{RUBY_FL_TAINT} */
 #define FL_SHAREABLE    RBIMPL_CAST((VALUE)RUBY_FL_SHAREABLE)            /**< @old{RUBY_FL_SHAREABLE} */
 #define FL_UNTRUSTED    RBIMPL_CAST((VALUE)RUBY_FL_UNTRUSTED)            /**< @old{RUBY_FL_UNTRUSTED} */
-#define FL_SEEN_OBJ_ID  RBIMPL_CAST((VALUE)RUBY_FL_SEEN_OBJ_ID)          /**< @old{RUBY_FL_SEEN_OBJ_ID} */
 #define FL_EXIVAR       RBIMPL_CAST((VALUE)RUBY_FL_EXIVAR)               /**< @old{RUBY_FL_EXIVAR} */
 #define FL_FREEZE       RBIMPL_CAST((VALUE)RUBY_FL_FREEZE)               /**< @old{RUBY_FL_FREEZE} */
 
@@ -111,13 +109,6 @@
 #define RB_OBJ_FREEZE_RAW    RB_OBJ_FREEZE_RAW
 #define RB_OBJ_FROZEN        RB_OBJ_FROZEN
 #define RB_OBJ_FROZEN_RAW    RB_OBJ_FROZEN_RAW
-#define RB_OBJ_INFECT        RB_OBJ_INFECT
-#define RB_OBJ_INFECT_RAW    RB_OBJ_INFECT_RAW
-#define RB_OBJ_TAINT         RB_OBJ_TAINT
-#define RB_OBJ_TAINTABLE     RB_OBJ_TAINTABLE
-#define RB_OBJ_TAINTED       RB_OBJ_TAINTED
-#define RB_OBJ_TAINTED_RAW   RB_OBJ_TAINTED_RAW
-#define RB_OBJ_TAINT_RAW     RB_OBJ_TAINT_RAW
 #define RB_OBJ_UNTRUST       RB_OBJ_TAINT
 #define RB_OBJ_UNTRUSTED     RB_OBJ_TAINTED
 /** @endcond */
@@ -207,12 +198,15 @@ ruby_fl_type {
     RUBY_FL_WB_PROTECTED = (1<<5),
 
     /**
-     * This flag  has something to do  with our garbage collector.   These days
-     * ruby  objects are  "generational".  There  are those  who are  young and
-     * those who are old.  Young objects are prone to die; monitored relatively
-     * extensively by  the garbage  collector.  OTOH old  objects tend  to live
-     * longer.  They are relatively rarely considered.  This flag is set when a
-     * object experienced promotion i.e. survived a garbage collection.
+     * Ruby objects are "generational".  There are young objects & old objects.
+     * Young objects are prone to die & monitored relatively extensively by the
+     * garbage collector.  Old objects tend to live longer & are monitored less
+     * frequently.  When an object survives a GC, its age is incremented.  When
+     * age is equal to RVALUE_OLD_AGE, the object becomes Old. This flag is set
+     * when an object becomes old, and is used by the write barrier to check if
+     * an old object should be considered for marking more frequently  - as old
+     * objects that have references added between major GCs need to be remarked
+     * to prevent the referred object being mistakenly swept.
      *
      * @internal
      *
@@ -220,41 +214,14 @@ ruby_fl_type {
      * 3rd parties.  It must be an implementation detail that they should never
      * know.  Might better be hidden.
      */
-    RUBY_FL_PROMOTED0    = (1<<5),
+    RUBY_FL_PROMOTED    = (1<<5),
 
     /**
-     * This flag  has something to do  with our garbage collector.   These days
-     * ruby  objects are  "generational".  There  are those  who are  young and
-     * those who are old.  Young objects are prone to die; monitored relatively
-     * extensively by  the garbage  collector.  OTOH old  objects tend  to live
-     * longer.  They are relatively rarely considered.  This flag is set when a
-     * object  experienced two  promotions  i.e.  survived garbage  collections
-     * twice.
+     * This flag meaning is type dependent, currently only used by T_DATA.
      *
      * @internal
-     *
-     * But honestly, @shyouhei  doesn't think this flag should  be visible from
-     * 3rd parties.  It must be an implementation detail that they should never
-     * know.  Might better be hidden.
      */
-    RUBY_FL_PROMOTED1    = (1<<6),
-
-    /**
-     * This flag  has something to do  with our garbage collector.   These days
-     * ruby  objects are  "generational".  There  are those  who are  young and
-     * those who are old.  Young objects are prone to die; monitored relatively
-     * extensively by  the garbage  collector.  OTOH old  objects tend  to live
-     * longer.  They are relatively rarely considered.  This flag is set when a
-     * object  experienced  promotions i.e.   survived  more  than one  garbage
-     * collections.
-     *
-     * @internal
-     *
-     * But honestly, @shyouhei  doesn't think this flag should  be visible from
-     * 3rd parties.  It must be an implementation detail that they should never
-     * know.  Might better be hidden.
-     */
-    RUBY_FL_PROMOTED     = RUBY_FL_PROMOTED0 | RUBY_FL_PROMOTED1,
+    RUBY_FL_USERPRIV0    = (1<<6),
 
     /**
      * This flag has  something to do with finalisers.  A  ruby object can have
@@ -283,7 +250,22 @@ ruby_fl_type {
 # pragma deprecated(RUBY_FL_TAINT)
 #endif
 
-                         = (1<<8),
+                         = 0,
+
+    /**
+     * @deprecated  This flag was an implementation detail that should never have
+     *              no been exposed. Exists  here for  backwards
+     *              compatibility only.  You can safely forget about it.
+     */
+    RUBY_FL_EXIVAR
+
+#if defined(RBIMPL_HAVE_ENUM_ATTRIBUTE)
+    RBIMPL_ATTR_DEPRECATED(("FL_EXIVAR is an outdated implementation detail, it shoudl be used."))
+#elif defined(_MSC_VER)
+# pragma deprecated(RUBY_FL_EXIVAR)
+#endif
+
+                         = 0,
 
     /**
      * This flag has something to do with Ractor.  Multiple Ractors run without
@@ -310,39 +292,21 @@ ruby_fl_type {
 # pragma deprecated(RUBY_FL_UNTRUSTED)
 #endif
 
-                         = (1<<8),
+                         = 0,
 
-    /**
-     * This flag has something to do with  object IDs.  Unlike in the old days,
-     * an object's object  ID (that a user can  query using `Object#object_id`)
-     * is no longer its physical address represented using Ruby level integers.
-     * It is  now a  monotonic-increasing integer  unrelated to  the underlying
-     * memory arrangement.  Object IDs are assigned when necessary; objects are
-     * born without one,  and will eventually have such  property when queried.
-     * The interpreter has to manage which one is which.  This is the flag that
-     * helps the  management.  Objects  with this  flag set  are the  ones with
-     * object IDs assigned.
-     *
-     * @internal
-     *
-     * But honestly, @shyouhei  doesn't think this flag should  be visible from
-     * 3rd parties.  It must be an implementation detail that they should never
-     * know.  Might better be hidden.
-     */
-    RUBY_FL_SEEN_OBJ_ID  = (1<<9),
+   /**
+    * This flag is no longer in use
+    *
+    * @internal
+    */
+    RUBY_FL_UNUSED9  = (1<<9),
 
-    /**
-     * This flag has something to do with instance variables.  3rd parties need
-     * not  know, but  there are  several ways  to store  an object's  instance
-     * variables.   Objects  with this  flag  use  so-called "generic"  backend
-     * storage.  This  distinction is purely an  implementation detail.  People
-     * need not be aware of this working behind-the-scene.
-     *
-     * @internal
-     *
-     * As of writing everything except ::RObject and RModule use this scheme.
-     */
-    RUBY_FL_EXIVAR       = (1<<10),
+   /**
+    * This flag is no longer in use
+    *
+    * @internal
+    */
+    RUBY_FL_UNUSED10 = (1<<10),
 
     /**
      * This flag has something to do with data immutability.  When this flag is
@@ -402,7 +366,7 @@ ruby_fl_type {
      * 3rd parties.  It must be an implementation detail that they should never
      * know.  Might better be hidden.
      */
-    RUBY_ELTS_SHARED  = RUBY_FL_USER2,
+    RUBY_ELTS_SHARED  = RUBY_FL_USER0,
 
     /**
      * This flag has something to do with an object's class.  There are kind of
@@ -427,7 +391,7 @@ ruby_fl_type {
      * 3rd parties.  It must be an implementation detail that they should never
      * know.  Might better be hidden.
      */
-    RUBY_FL_SINGLETON = RUBY_FL_USER0,
+    RUBY_FL_SINGLETON = RUBY_FL_USER1,
 };
 
 enum {
@@ -444,7 +408,7 @@ enum {
 # pragma deprecated(RUBY_FL_DUPPED)
 #endif
 
-    = (int)RUBY_T_MASK | (int)RUBY_FL_EXIVAR
+    = (int)RUBY_T_MASK
 };
 
 #undef RBIMPL_HAVE_ENUM_ATTRIBUTE
@@ -478,10 +442,8 @@ RB_FL_ABLE(VALUE obj)
     if (RB_SPECIAL_CONST_P(obj)) {
         return false;
     }
-    else if (RB_TYPE_P(obj, RUBY_T_NODE)) {
-        return false;
-    }
     else {
+        RBIMPL_ASSERT_OR_ASSUME(!RB_TYPE_P(obj, RUBY_T_NODE));
         return true;
     }
 }
@@ -489,7 +451,7 @@ RB_FL_ABLE(VALUE obj)
 RBIMPL_ATTR_PURE_UNLESS_DEBUG()
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an implenentation detail of  RB_FL_TEST().  3rd parties need not use
+ * This is an implementation detail of  RB_FL_TEST().  3rd parties need not use
  * this.  Just always use RB_FL_TEST().
  *
  * @param[in]  obj    Object in question.
@@ -537,7 +499,7 @@ RB_FL_TEST(VALUE obj, VALUE flags)
 RBIMPL_ATTR_PURE_UNLESS_DEBUG()
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an  implenentation detail of RB_FL_ANY().  3rd parties  need not use
+ * This is an  implementation detail of RB_FL_ANY().  3rd parties  need not use
  * this.  Just always use RB_FL_ANY().
  *
  * @param[in]  obj    Object in question.
@@ -571,7 +533,7 @@ RB_FL_ANY(VALUE obj, VALUE flags)
 RBIMPL_ATTR_PURE_UNLESS_DEBUG()
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an  implenentation detail of RB_FL_ALL().  3rd parties  need not use
+ * This is an  implementation detail of RB_FL_ALL().  3rd parties  need not use
  * this.  Just always use RB_FL_ALL().
  *
  * @param[in]  obj    Object in question.
@@ -607,7 +569,7 @@ RBIMPL_ATTR_ARTIFICIAL()
 /**
  * @private
  *
- * This is an  implenentation detail of RB_FL_SET().  3rd parties  need not use
+ * This is an  implementation detail of RB_FL_SET().  3rd parties  need not use
  * this.  Just always use RB_FL_SET().
  *
  * @param[out]  obj    Object in question.
@@ -627,7 +589,7 @@ rbimpl_fl_set_raw_raw(struct RBasic *obj, VALUE flags)
 
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an  implenentation detail of RB_FL_SET().  3rd parties  need not use
+ * This is an  implementation detail of RB_FL_SET().  3rd parties  need not use
  * this.  Just always use RB_FL_SET().
  *
  * @param[out]  obj    Object in question.
@@ -667,7 +629,7 @@ RBIMPL_ATTR_ARTIFICIAL()
 /**
  * @private
  *
- * This is an implenentation detail of RB_FL_UNSET().  3rd parties need not use
+ * This is an implementation detail of RB_FL_UNSET().  3rd parties need not use
  * this.  Just always use RB_FL_UNSET().
  *
  * @param[out]  obj    Object in question.
@@ -687,7 +649,7 @@ rbimpl_fl_unset_raw_raw(struct RBasic *obj, VALUE flags)
 
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an implenentation detail of RB_FL_UNSET().  3rd parties need not use
+ * This is an implementation detail of RB_FL_UNSET().  3rd parties need not use
  * this.  Just always use RB_FL_UNSET().
  *
  * @param[out]  obj    Object in question.
@@ -722,7 +684,7 @@ RBIMPL_ATTR_ARTIFICIAL()
 /**
  * @private
  *
- * This is an  implenentation detail of RB_FL_REVERSE().  3rd  parties need not
+ * This is an  implementation detail of RB_FL_REVERSE().  3rd  parties need not
  * use this.  Just always use RB_FL_REVERSE().
  *
  * @param[out]  obj    Object in question.
@@ -742,7 +704,7 @@ rbimpl_fl_reverse_raw_raw(struct RBasic *obj, VALUE flags)
 
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an  implenentation detail of RB_FL_REVERSE().  3rd  parties need not
+ * This is an  implementation detail of RB_FL_REVERSE().  3rd  parties need not
  * use this.  Just always use RB_FL_REVERSE().
  *
  * @param[out]  obj    Object in question.
@@ -787,6 +749,7 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline bool
 RB_OBJ_TAINTABLE(VALUE obj)
 {
+    (void)obj;
     return false;
 }
 
@@ -804,6 +767,7 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline VALUE
 RB_OBJ_TAINTED_RAW(VALUE obj)
 {
+    (void)obj;
     return false;
 }
 
@@ -821,6 +785,7 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline bool
 RB_OBJ_TAINTED(VALUE obj)
 {
+    (void)obj;
     return false;
 }
 
@@ -836,6 +801,7 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline void
 RB_OBJ_TAINT_RAW(VALUE obj)
 {
+    (void)obj;
     return;
 }
 
@@ -851,6 +817,7 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline void
 RB_OBJ_TAINT(VALUE obj)
 {
+    (void)obj;
     return;
 }
 
@@ -867,6 +834,8 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline void
 RB_OBJ_INFECT_RAW(VALUE dst, VALUE src)
 {
+    (void)dst;
+    (void)src;
     return;
 }
 
@@ -883,13 +852,15 @@ RBIMPL_ATTR_DEPRECATED(("taintedness turned out to be a wrong idea."))
 static inline void
 RB_OBJ_INFECT(VALUE dst, VALUE src)
 {
+    (void)dst;
+    (void)src;
     return;
 }
 
 RBIMPL_ATTR_PURE_UNLESS_DEBUG()
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an  implenentation detail of RB_OBJ_FROZEN().  3rd  parties need not
+ * This is an  implementation detail of RB_OBJ_FROZEN().  3rd  parties need not
  * use this.  Just always use RB_OBJ_FROZEN().
  *
  * @param[in]  obj             Object in question.
@@ -928,9 +899,21 @@ RB_OBJ_FROZEN(VALUE obj)
     }
 }
 
+RUBY_SYMBOL_EXPORT_BEGIN
+/**
+ * Prevents further modifications to the given object.  ::rb_eFrozenError shall
+ * be raised if modification is attempted.
+ *
+ * @param[out]  x               Object in question.
+ * @exception   rb_eNoMemError  Failed   to   allocate memory  for  the  frozen
+ *                              representation of the object.
+ */
+void rb_obj_freeze_inline(VALUE obj);
+RUBY_SYMBOL_EXPORT_END
+
 RBIMPL_ATTR_ARTIFICIAL()
 /**
- * This is an  implenentation detail of RB_OBJ_FREEZE().  3rd  parties need not
+ * This is an  implementation detail of RB_OBJ_FREEZE().  3rd  parties need not
  * use this.  Just always use RB_OBJ_FREEZE().
  *
  * @param[out]  obj  Object in question.
@@ -938,24 +921,7 @@ RBIMPL_ATTR_ARTIFICIAL()
 static inline void
 RB_OBJ_FREEZE_RAW(VALUE obj)
 {
-    RB_FL_SET_RAW(obj, RUBY_FL_FREEZE);
-}
-
-/**
- * Prevents further modifications to the given object.  ::rb_eFrozenError shall
- * be raised if modification is attempted.
- *
- * @param[out]  x  Object in question.
- */
-static inline void
-rb_obj_freeze_inline(VALUE x)
-{
-    if (RB_FL_ABLE(x)) {
-        RB_OBJ_FREEZE_RAW(x);
-        if (RBASIC_CLASS(x) && !(RBASIC(x)->flags & RUBY_FL_SINGLETON)) {
-            rb_freeze_singleton_class(x);
-        }
-    }
+    rb_obj_freeze_inline(obj);
 }
 
 #endif /* RBIMPL_FL_TYPE_H */
