@@ -1,6 +1,6 @@
 /**********************************************************************
 
-  vm_args.c - process method call arguments.
+  vm_args.c - process method call arguments. Included into vm.c.
 
   $Author$
 
@@ -8,9 +8,9 @@
 
 **********************************************************************/
 
-NORETURN(static void raise_argument_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const VALUE exc));
-NORETURN(static void argument_arity_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const int miss_argc, const int min_argc, const int max_argc));
-NORETURN(static void argument_kw_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const char *error, const VALUE keys));
+NORETURN(static void raise_argument_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const rb_callable_method_entry_t *cme, const VALUE exc));
+NORETURN(static void argument_arity_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const rb_callable_method_entry_t *cme, const int miss_argc, const int min_argc, const int max_argc));
+NORETURN(static void argument_kw_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const rb_callable_method_entry_t *cme, const char *error, const VALUE keys));
 VALUE rb_keyword_error_new(const char *error, VALUE keys); /* class.c */
 static VALUE method_missing(rb_execution_context_t *ec, VALUE obj, ID id, int argc, const VALUE *argv,
                             enum method_missing_reason call_status, int kw_splat);
@@ -321,7 +321,7 @@ args_setup_kw_parameters_lookup(const ID key, VALUE *ptr, const VALUE *const pas
 #define KW_SPECIFIED_BITS_MAX (32-1) /* TODO: 32 -> Fixnum's max bits */
 
 static void
-args_setup_kw_parameters(rb_execution_context_t *const ec, const rb_iseq_t *const iseq,
+args_setup_kw_parameters(rb_execution_context_t *const ec, const rb_iseq_t *const iseq, const rb_callable_method_entry_t *cme,
                          VALUE *const passed_values, const int passed_keyword_len, const VALUE *const passed_keywords,
                          VALUE *const locals)
 {
@@ -345,7 +345,7 @@ args_setup_kw_parameters(rb_execution_context_t *const ec, const rb_iseq_t *cons
         }
     }
 
-    if (missing) argument_kw_error(ec, iseq, "missing", missing);
+    if (missing) argument_kw_error(ec, iseq, cme, "missing", missing);
 
     for (di=0; i<key_num; i++, di++) {
         if (args_setup_kw_parameters_lookup(acceptable_keywords[i], &locals[i], passed_keywords, passed_values, passed_keyword_len)) {
@@ -386,7 +386,7 @@ args_setup_kw_parameters(rb_execution_context_t *const ec, const rb_iseq_t *cons
     else {
         if (found != passed_keyword_len) {
             VALUE keys = make_unknown_kw_hash(passed_keywords, passed_keyword_len, passed_values);
-            argument_kw_error(ec, iseq, "unknown", keys);
+            argument_kw_error(ec, iseq, cme, "unknown", keys);
         }
     }
 
@@ -397,7 +397,7 @@ args_setup_kw_parameters(rb_execution_context_t *const ec, const rb_iseq_t *cons
 }
 
 static void
-args_setup_kw_parameters_from_kwsplat(rb_execution_context_t *const ec, const rb_iseq_t *const iseq,
+args_setup_kw_parameters_from_kwsplat(rb_execution_context_t *const ec, const rb_iseq_t *const iseq, const rb_callable_method_entry_t *cme,
                          VALUE keyword_hash, VALUE *const locals, bool remove_hash_value)
 {
     const ID *acceptable_keywords = ISEQ_BODY(iseq)->param.keyword->table;
@@ -430,7 +430,7 @@ args_setup_kw_parameters_from_kwsplat(rb_execution_context_t *const ec, const rb
         }
     }
 
-    if (missing) argument_kw_error(ec, iseq, "missing", missing);
+    if (missing) argument_kw_error(ec, iseq, cme, "missing", missing);
 
     for (di=0; i<key_num; i++, di++) {
         VALUE key = ID2SYM(acceptable_keywords[i]);
@@ -485,11 +485,11 @@ args_setup_kw_parameters_from_kwsplat(rb_execution_context_t *const ec, const rb
                  * This is simpler than writing code to check which entries in the hash do not match.
                  * This will raise an exception, so the additional performance impact shouldn't be material.
                  */
-                args_setup_kw_parameters_from_kwsplat(ec, iseq, rb_hash_dup(keyword_hash), locals, true);
+                args_setup_kw_parameters_from_kwsplat(ec, iseq, cme, rb_hash_dup(keyword_hash), locals, true);
             }
         }
         else if (!RHASH_EMPTY_P(keyword_hash)) {
-            argument_kw_error(ec, iseq, "unknown", rb_hash_keys(keyword_hash));
+            argument_kw_error(ec, iseq, cme, "unknown", rb_hash_keys(keyword_hash));
         }
     }
 
@@ -579,7 +579,7 @@ flatten_rest_args(rb_execution_context_t * const ec, struct args_info *args, VAL
     args->argc += rest_len;
     if (rest_len) {
         CHECK_VM_STACK_OVERFLOW(ec->cfp, rest_len+1);
-        for (i, j=0; rest_len > 0; rest_len--, i++, j++) {
+        for (j=0; rest_len > 0; rest_len--, i++, j++) {
             locals[i] = argv[j];
         }
     }
@@ -607,6 +607,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     VALUE splat_flagged_keyword_hash = 0;
     VALUE converted_keyword_hash = 0;
     VALUE rest_last = 0;
+    const rb_callable_method_entry_t *cme = calling->cc ? vm_cc_cme(calling->cc) : NULL;
 
     vm_check_canary(ec, orig_sp);
     /*
@@ -861,7 +862,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
             args_extend(args, min_argc);
         }
         else {
-            argument_arity_error(ec, iseq, given_argc, min_argc, max_argc);
+            argument_arity_error(ec, iseq, cme, given_argc, min_argc, max_argc);
         }
     }
 
@@ -872,7 +873,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
             given_argc = max_argc;
         }
         else {
-            argument_arity_error(ec, iseq, given_argc, min_argc, max_argc);
+            argument_arity_error(ec, iseq, cme, given_argc, min_argc, max_argc);
         }
     }
 
@@ -894,16 +895,22 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     }
 
     if (ISEQ_BODY(iseq)->param.flags.has_rest) {
-        args_setup_rest_parameter(args, locals + ISEQ_BODY(iseq)->param.rest_start);
-        VALUE ary = *(locals + ISEQ_BODY(iseq)->param.rest_start);
-        VALUE index = RARRAY_LEN(ary) - 1;
-        if (splat_flagged_keyword_hash &&
-            !ISEQ_BODY(iseq)->param.flags.ruby2_keywords &&
-            !ISEQ_BODY(iseq)->param.flags.has_kw &&
-            !ISEQ_BODY(iseq)->param.flags.has_kwrest &&
-            RARRAY_AREF(ary, index) == splat_flagged_keyword_hash) {
-            ((struct RHash *)rest_last)->basic.flags &= ~RHASH_PASS_AS_KEYWORDS;
-            RARRAY_ASET(ary, index, rest_last);
+        if (UNLIKELY(ISEQ_BODY(iseq)->param.flags.anon_rest && args->argc == 0 && !args->rest && !ISEQ_BODY(iseq)->param.flags.has_post)) {
+           *(locals + ISEQ_BODY(iseq)->param.rest_start) = args->rest = rb_cArray_empty_frozen;
+           args->rest_index = 0;
+        }
+        else {
+            args_setup_rest_parameter(args, locals + ISEQ_BODY(iseq)->param.rest_start);
+            VALUE ary = *(locals + ISEQ_BODY(iseq)->param.rest_start);
+            VALUE index = RARRAY_LEN(ary) - 1;
+            if (splat_flagged_keyword_hash &&
+                !ISEQ_BODY(iseq)->param.flags.ruby2_keywords &&
+                !ISEQ_BODY(iseq)->param.flags.has_kw &&
+                !ISEQ_BODY(iseq)->param.flags.has_kwrest &&
+                RARRAY_AREF(ary, index) == splat_flagged_keyword_hash) {
+                ((struct RHash *)rest_last)->basic.flags &= ~RHASH_PASS_AS_KEYWORDS;
+                RARRAY_ASET(ary, index, rest_last);
+            }
         }
     }
 
@@ -912,7 +919,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 
         if (args->kw_argv != NULL) {
             const struct rb_callinfo_kwarg *kw_arg = args->kw_arg;
-            args_setup_kw_parameters(ec, iseq, args->kw_argv, kw_arg->keyword_len, kw_arg->keywords, klocals);
+            args_setup_kw_parameters(ec, iseq, cme, args->kw_argv, kw_arg->keyword_len, kw_arg->keywords, klocals);
         }
         else if (!NIL_P(keyword_hash)) {
             bool remove_hash_value = false;
@@ -920,7 +927,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
                 keyword_hash = check_kwrestarg(keyword_hash, &kw_flag);
                 remove_hash_value = true;
             }
-            args_setup_kw_parameters_from_kwsplat(ec, iseq, keyword_hash, klocals, remove_hash_value);
+            args_setup_kw_parameters_from_kwsplat(ec, iseq, cme, keyword_hash, klocals, remove_hash_value);
         }
         else {
 #if VM_CHECK_MODE > 0
@@ -935,7 +942,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
                 VM_ASSERT(args_argc(args) == 1);
             }
 #endif
-            args_setup_kw_parameters(ec, iseq, NULL, 0, NULL, klocals);
+            args_setup_kw_parameters(ec, iseq, cme, NULL, 0, NULL, klocals);
         }
     }
     else if (ISEQ_BODY(iseq)->param.flags.has_kwrest) {
@@ -943,7 +950,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
             kw_flag, ISEQ_BODY(iseq)->param.flags.anon_kwrest);
     }
     else if (!NIL_P(keyword_hash) && RHASH_SIZE(keyword_hash) > 0 && arg_setup_type == arg_setup_method) {
-        argument_kw_error(ec, iseq, "unknown", rb_hash_keys(keyword_hash));
+        argument_kw_error(ec, iseq, cme, "unknown", rb_hash_keys(keyword_hash));
     }
 
     if (ISEQ_BODY(iseq)->param.flags.has_block) {
@@ -969,17 +976,16 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 }
 
 static void
-raise_argument_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const VALUE exc)
+raise_argument_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const rb_callable_method_entry_t *cme, const VALUE exc)
 {
     VALUE at;
 
     if (iseq) {
         vm_push_frame(ec, iseq, VM_FRAME_MAGIC_DUMMY | VM_ENV_FLAG_LOCAL, Qnil /* self */,
-                      VM_BLOCK_HANDLER_NONE /* specval*/, Qfalse /* me or cref */,
+                      VM_BLOCK_HANDLER_NONE /* specval*/, (VALUE) cme /* me or cref */,
                       ISEQ_BODY(iseq)->iseq_encoded,
                       ec->cfp->sp, 0, 0 /* stack_max */);
         at = rb_ec_backtrace_object(ec);
-        rb_backtrace_use_iseq_first_lineno_for_last_location(at);
         rb_vm_pop_frame(ec);
     }
     else {
@@ -992,7 +998,7 @@ raise_argument_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const VA
 }
 
 static void
-argument_arity_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const int miss_argc, const int min_argc, const int max_argc)
+argument_arity_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const rb_callable_method_entry_t *cme, const int miss_argc, const int min_argc, const int max_argc)
 {
     VALUE exc = rb_arity_error_new(miss_argc, min_argc, max_argc);
     if (ISEQ_BODY(iseq)->param.flags.has_kw) {
@@ -1013,13 +1019,13 @@ argument_arity_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const in
             RSTRING_PTR(mesg)[RSTRING_LEN(mesg)-1] = ')';
         }
     }
-    raise_argument_error(ec, iseq, exc);
+    raise_argument_error(ec, iseq, cme, exc);
 }
 
 static void
-argument_kw_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const char *error, const VALUE keys)
+argument_kw_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const rb_callable_method_entry_t *cme, const char *error, const VALUE keys)
 {
-    raise_argument_error(ec, iseq, rb_keyword_error_new(error, keys));
+    raise_argument_error(ec, iseq, cme, rb_keyword_error_new(error, keys));
 }
 
 static VALUE
@@ -1039,9 +1045,17 @@ vm_to_proc(VALUE proc)
         }
 
         if (NIL_P(b) || !rb_obj_is_proc(b)) {
-            rb_raise(rb_eTypeError,
-                     "wrong argument type %s (expected Proc)",
-                     rb_obj_classname(proc));
+            if (me) {
+                VALUE cname = rb_obj_class(proc);
+                rb_raise(rb_eTypeError,
+                         "can't convert %"PRIsVALUE" to Proc (%"PRIsVALUE"#to_proc gives %"PRIsVALUE")",
+                         cname, cname, rb_obj_class(b));
+            }
+            else {
+                rb_raise(rb_eTypeError,
+                         "no implicit conversion of %s into Proc",
+                         rb_obj_classname(proc));
+            }
         }
         return b;
     }
@@ -1177,7 +1191,8 @@ vm_caller_setup_fwd_args(const rb_execution_context_t *ec, rb_control_frame_t *r
 
     *adjusted_ci = VM_CI_ON_STACK(
             site_mid,
-            (caller_flag | (site_flag & (VM_CALL_FCALL | VM_CALL_FORWARDING))),
+            ((caller_flag & ~(VM_CALL_ARGS_SIMPLE | VM_CALL_FCALL)) |
+             (site_flag & (VM_CALL_FCALL | VM_CALL_FORWARDING))),
             site_argc + caller_argc,
             kw
             );
